# -*- coding: utf-8 -*-
"""
Utilities for the CLI functions.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import os.path as path
import re

import click

from .. import Crumb


# different context options
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
UNKNOWN_OPTIONS = dict(allow_extra_args=True,
                       ignore_unknown_options=True)


# specification of existing ParamTypes
ExistingDirPath  = click.Path(exists=True, file_okay=False, resolve_path=True)
ExistingFilePath = click.Path(exists=True,  dir_okay=False, resolve_path=True)
UnexistingFilePath = click.Path(dir_okay=False, resolve_path=True)


# validators
def check_not_none(ctx, param, value):
    if value is None:
        raise click.BadParameter('got {}.'.format(value))
    return value


# declare custom click.ParamType
class RegularExpression(click.ParamType):
    name = 'regex'

    def convert(self, value, param, ctx):
        try:
            rex = re.compile(value, re.IGNORECASE)
        except ValueError:
            self.fail('%s is not a valid regular expression.' % value, param, ctx)
        else:
            return rex


class CrumbPath(click.ParamType):
    name = 'crumb'

    def convert(self, value, param, ctx):
        try:
            cr = Crumb(path.expanduser(value), ignore_list=['.*'])
        except ValueError:
            self.fail('%s is not a valid crumb path.' % value, param, ctx)
        else:
            return cr

# other utilities
def echo_list(alist):
    for i in alist:
        click.echo(i)


def _print_values_map_as_csv(list_of_lists):
    for values in list_of_lists:
        click.echo(','.join([item[1] for item in values]))



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Utilities to make crumbs
"""
import os
import re
import os.path as op
import fnmatch
import operator
import itertools
import shutil
from   collections import Mapping, defaultdict, OrderedDict
from   copy        import deepcopy
from   functools   import partial, reduce

from   six import string_types

from ._utils import _check_is_subset, _is_crumb_arg


def rm_dups(lst):
    """Return a sorted lst of non-duplicated elements from `lst`.

    Parameters
    ----------
    lst: sequence of any

    Returns
    -------
    fslst:
        Filtered and sorted `lst` with non duplicated elements of `lst`.
    """
    return sorted(list(set(lst)))


def remove_ignored(ignore, strs):
    """Remove from `strs` the matches to the `fnmatch` (glob) patterns and
    return the result in a list."""
    nustrs = deepcopy(strs)
    for ign in ignore:
        nustrs = (item for item in nustrs if not fnmatch.fnmatch(item, ign))

    return nustrs


def fnmatch_filter(pattern, items, *args):
    """Return the items from `items` that match the fnmatch expression in
    `pattern`.
    Parameters
    ----------
    pattern: str
        Regular expression

    items: list of str
        The items to be checked

    args: ignored

    Returns
    -------
    matches: list of str
        Matched items
    """
    return (item for item in items if fnmatch.fnmatch(item, pattern))


def regex_match_filter(pattern, items, *args):
    """Return the items from `items` that match the regular expression in
    `pattern`.
    Parameters
    ----------
    pattern: str
        Regular expression

    items: list of str
        The items to be checked

    args: re.compile arguments

    Returns
    -------
    matches: list of str
        Matched items
    """
    test = re.compile(pattern, *args)
    return (s for s in items if test.match(s))


def list_children(path, just_dirs=False):
    """Return the immediate elements (files and folders) in `path`.
    Parameters
    ----------
    path: str

    just_dirs: bool
        If True will return only folders.

    ignore: sequence of str
        Sequence of glob patterns to ignore from the listing.

    re: str
        Regular expression that the result items must match.

    Returns
    -------
    paths: list of str
    """
    if not op.exists(path):
        raise IOError("Expected an existing path, but could not"
                      " find {}.".format(path))

    if op.isfile(path):
        if just_dirs:
            vals = []
        else:
            vals = [path]
    else:
        if just_dirs: # this means we have to list only folders
            vals = [d for d in os.listdir(path) if op.isdir(op.join(path, d))]
        else:   # this means we have to list files
            vals = os.listdir(path)

    return vals


def list_subpaths(path, just_dirs=False, ignore=None, pattern=None,
                  filter_func=fnmatch_filter, filter_args=None):
    """Return the immediate elements (files and folders) within `path`.
    Parameters
    ----------
    path: str

    just_dirs: bool
        If True will return only folders.

    ignore: sequence of str
        Sequence of glob patterns to ignore from the listing.

    pattern: str
        Regular expression that the result items must match.

    filter_func: func
        The function to match the patterns.
        Must have as arguments: (pattern, paths) and return
        a subset of paths.

    filter_args: filter func arguments
        Arguments for the filter function.

    Returns
    -------
    paths: list of str
    """
    paths = list_children(path, just_dirs=just_dirs)

    if ignore and ignore is not None:
        paths = remove_ignored(ignore, paths)

    if pattern and pattern is not None:
        if filter_args is None:
            filter_args = ()

        paths = filter_func(pattern, paths, *filter_args)

    return paths


def list_intersection(list1, list2):
    """Return a list of elements that are the intersection between the set of
    elements of `list1` and `list2`·
    This will keep the same order of the elements in `list1`.
    """
    return (arg_name for arg_name in list1 if arg_name in list2)


def _get_matching_items(list1, list2, items=None):
    """If `items` is None, Return a list of items that are in
    `list1` and `list2`. Otherwise will return the elements of `items` if
    they are in both lists.
    Keep the order in `list1` or in `items`.

    Returns
    -------
    arg_names: list
        The matching items.

    Raises
    ------
    ValueError:
        If an element of items does not exists in either `list1` or `list2`.
    """
    if not items or items is None:
        arg_names = list_intersection(list1, list2)
    else:
        try:
            _check_is_subset(items, list1)
            _check_is_subset(items, list2)
        except KeyError:
            arg_names = []
        except:
            raise
        else:
            arg_names = items

    return arg_names


def joint_value_map(crumb, arg_names, check_exists=True):
    """Return a list of tuples of crumb argument values of the given
    `arg_names`.
    Parameters
    ----------
    arg_name: str

    check_exists: bool
        If True will return only a values_map with sets of crumb arguments that
        fill a crumb to an existing path.
        Otherwise it won't check if they exist and return all possible
        combinations.

    Returns
    -------
    values_map: list of lists of 2-tuples
        I call values_map what is called `record` in pandas. It is a list of
        lists of 2-tuples, where each 2-tuple has the
        shape (arg_name, arg_value).
    """
    values_map = []
    for arg_name in arg_names:
        values_map.append(list((arg_name, arg_value)
                          for arg_value in crumb[arg_name]))

    if len(arg_names) == 1:
        return [(i, ) for i in values_map[0]]
    else:
        if not check_exists:
            values_map_checked = values_map[:]
        else:
            args_crumbs = [(args, crumb.replace(**dict(args)))
                           for args in set(itertools.product(*values_map))]

            values_map_checked = [args for args, cr in args_crumbs
                                  if cr.exists()]

    return sorted(values_map_checked)


def intersection(crumb1, crumb2, on=None):
    """Return an 'inner join' of both given Crumbs, i.e., will return a list of
    Crumbs with common values for the common arguments of both crumbs.

    If `on` is None, will use all the common arguments names of both crumbs.
    Otherwise will use only the elements of `on`. All its items must be in
    both crumbs.

    Returns
    -------
    arg_names: list
        The matching items.

    Parameters
    ----------
    crumb1: hansel.Crumb

    crumb2: hansel.Crumb

    on: str or list of str
        Crumb argument names common to both input crumbs.

    Raises
    ------
    ValueError:
        If an element of `on` does not exists in either `list1` or `list2`.

    KeyError:
        If the result is empty.

    Returns
    -------
    inner_join: list[hansel.Crumb]

    Notes
    -----
    Use with care, ideally the argument matches should be in the same order in
    both crumbs.

    Both crumbs must have at least one matching identifier argument and one
    of those must be the one in `id_colname`.
    """
    if isinstance(on, string_types):
        on = [on]

    arg_names = list(_get_matching_items(list(crumb1.all_args()),
                                         list(crumb2.all_args()),
                                         items=on))

    if not arg_names:
        raise KeyError("Could not find matching arguments between "
                       "{} and  {} limited by {}.".format(list(crumb1.all_args()),
                                                          list(crumb2.all_args()),
                                                          on))

    maps1 = joint_value_map(crumb1, arg_names, check_exists=True)
    maps2 = joint_value_map(crumb2, arg_names, check_exists=True)

    intersect = set(maps1) & (set(maps2))

    return sorted(list(intersect))


def difference(crumb1, crumb2, on=None):
    """Return the difference `crumb1` - `crumb2`, i.e., will return a list of
    Crumbs that are in `crumb1` but not in `crumb2`.

    If `on` is None, will use all the common arguments names of both crumbs.
    Otherwise will use only the elements of `on`. All its items must be in
    both crumbs.

    Returns
    -------
    arg_names: list
        The matching items.

    Parameters
    ----------
    crumb1: hansel.Crumb

    crumb2: hansel.Crumb

    on: str or list of str
        Crumb argument names common to both input crumbs.

    Raises
    ------
    ValueError:
        If an element of `on` does not exists in either `list1` or `list2`.

    KeyError:
        If the result is empty.

    Returns
    -------
    inner_join: list[hansel.Crumb]

    Notes
    -----
    Use with care, ideally the argument matches should be in the same order in
    both crumbs.

    Both crumbs must have at least one matching identifier argument and one
    of those must be the one in `id_colname`.
    """
    if isinstance(on, string_types):
        on = [on]

    arg_names = list(_get_matching_items(list(crumb1.all_args()),
                                         list(crumb2.all_args()),
                                         items=on))

    if not arg_names:
        raise KeyError("Could not find matching arguments between "
                       "{} and  {} limited by {}.".format(list(crumb1.all_args()),
                                                          list(crumb2.all_args()),
                                                          on))

    maps1 = joint_value_map(crumb1, arg_names, check_exists=True)
    maps2 = joint_value_map(crumb2, arg_names, check_exists=True)

    diff = set(maps1).difference(set(maps2))

    return sorted(list(diff))


def valuesmap_to_dict(values_map):
    """Converts a values_map or records type (a list of list of 2-tuple with
    shape '(arg_name, arg_value)') to a dictionary of lists of values where the
    keys are the arg_names.
    Parameters
    ----------
    values_map: list of list of 2-tuple of str

    Returns
    -------
    adict: dict
        The dictionary with the values in `values_map` in the form of a
        dictionary.

    Raises
    ------
    IndexError
        If the list_of_dicts is empty or can't be indexed.

    KeyError
        If any list inside the `values_map` doesn't have all the keys in the
        first dict.
    """
    return append_dict_values([OrderedDict(rec) for rec in values_map])


def append_dict_values(list_of_dicts, keys=None):
    """Return a dict of lists from a list of dicts with the same keys as the
    internal dicts.
    For each dict in list_of_dicts with look for the values of the given keys
    and append it to the output dict.

    Parameters
    ----------
    list_of_dicts: list of dicts
        The first dict in this list will be used as reference for the key names
        of all the other dicts.

    keys: list of str
        List of keys to create in the output dict
        If None will use all keys in the first element of list_of_dicts
    Returns
    -------
    DefaultOrderedDict of lists

    Raises
    ------
    IndexError
        If the list_of_dicts is empty or can't be indexed.

    KeyError
        If any dict inside the `list_of_dicts` doesn't have all the keys in the
        first dict.
    """
    if keys is None:
        try:
            keys = list(list_of_dicts[0].keys())
        except IndexError:
            raise IndexError('Could not get the first element of the list.')

    dict_of_lists = defaultdict(list)
    for d in list_of_dicts:
        for k in keys:
            dict_of_lists[k].append(d[k])
    return dict_of_lists


def copy_args(src_crumb, dst_crumb):
    """Will copy the argument values of `src_crumb` to the open arguments of
    `dst_crumb`.
    """
    for arg_name in dst_crumb.open_args():
        dst_crumb[arg_name] = src_crumb[arg_name][0]


def _remove_if_ok_and_exists(path, exist_ok):
    if not exist_ok and op.exists(path):
        raise FileExistsError('Path {} already exists.'.format(path))
    elif op.exists(path):
        os.remove(path)


def copy_all_files(src_path, dst_path, exist_ok=True, verbose=False):
    """Will copy everything from `src_path` to `dst_path`.
    Both can be a folder path or a file path.
    """
    copy_func = shutil.copy2
    if verbose:
        print("Copying {} -> {}".format(src_path, dst_path))

    if op.isdir(src_path):
        if exist_ok:
            shutil.rmtree(dst_path)

        shutil.copytree(src_path, dst_path, copy_function=copy_func)
    elif op.isfile(src_path):
        os.makedirs(op.dirname(dst_path), exist_ok=exist_ok)
        try:
            copy_func(src_path, dst_path, follow_symlinks=True)
        except shutil.SameFileError:
            os.remove(dst_path)
            copy_func(src_path, dst_path, follow_symlinks=True)


def link_all_files(src_path, dst_path, exist_ok=True, verbose=False):
    """Make link from src_path to dst_path."""
    if not op.isabs(src_path):
        src_path = op.relpath(src_path, op.dirname(dst_path))

    if verbose:
        print("Linking {} -> {}".format(src_path, dst_path))

    os.makedirs(op.dirname(dst_path), exist_ok=True)

    _remove_if_ok_and_exists(dst_path, exist_ok=exist_ok)
    os.symlink(src_path, dst_path)


def _crumb_fill_dst(src_crumb, dst_crumb):
    """ Will list `src_crumb` and copy the resulting item arguments into
    `dst_crumb`.
    All the defined arguments of `src_crumb.ls()[0]` must define `dst_crumb`
    entirely and create a path to a file or folder.
    """
    for src in src_crumb.ls():
        dst = dst_crumb.copy()
        copy_args(src, dst)
        if dst.has_crumbs():
            raise AttributeError("Destination crumb still has open arguments, "
                                 "expected to fill it. Got {}.".format(str(dst)))
        yield src, dst


def crumb_copy(src_crumb, dst_crumb, exist_ok=False, verbose=False):
    """Will copy the content of `src_crumb` into `dst_crumb` folder.
    For this `src_crumb` and `dst_crumb` must have similar set of argument
    names.
    All the defined arguments of `src_crumb.ls()[0]` must define `dst_crumb`
    entirely and create a path to a file or folder.
    """
    for src, dst in _crumb_fill_dst(src_crumb, dst_crumb):
        copy_all_files(src.path, dst.path, exist_ok=exist_ok, verbose=verbose)


def crumb_link(src_crumb, dst_crumb, exist_ok=False, verbose=False):
    """Will link the content of `src_crumb` into `dst_crumb` folder.
    For this `src_crumb` and `dst_crumb` must have similar set of argument
    names.
    All the defined arguments of `src_crumb.ls()[0]` must define `dst_crumb`
    entirely and create a path to a file or folder.
    It will create the folder structure in the base of `dst_crumb` and link
    exclusively the leaf nodes.
    """
    for src, dst in _crumb_fill_dst(src_crumb, dst_crumb):
        link_all_files(src.path, dst.path, exist_ok=exist_ok, verbose=verbose)


def groupby_pattern(crumb, arg_name, groups):
    """Return a dictionary with the matches of `groups` values in the
    crumb argument `arg_name` in `crumb`.

    Parameters
    ----------
    crumb: Crumb
        Crumb to the folder tree.

    arg_name: str
        Name of the crumb argument in `crumb` that must be matched with the
        values of the `groups` dict.

    groups: dict[str]->str
        A dict where the keys are group names and the values are regular
        expressions (fnmatch xor re).

    Returns
    -------
    grouped: dict[str] -> list[Crumb]
        Map of paths from groups to the corresponding path matches.
    """
    if arg_name not in crumb:
        raise KeyError('Crumb {} has no argument {}.'.format(crumb, arg_name))

    paths_matched = set()
    mods = defaultdict(list)
    for mod_name, pattern in groups.items():
        crumb.set_pattern(arg_name, pattern)
        paths = crumb.ls(arg_name)
        if paths:
            mods[mod_name] = paths
            paths_matched = paths_matched.union(paths)

        crumb.clear_pattern(arg_name)

    return mods


def is_crumb_arg(crumb_arg):
    """ Return True if `crumb_arg` is a well formed crumb argument, i.e.,
    is a string that starts with `start_sym` and ends with `end_sym`.
    False otherwise.
    """
    return _is_crumb_arg(crumb_arg)


class ParameterGrid(object):
    """
    Picked from sklearn: https://github.com/scikit-learn/scikit-learn

    Grid of parameters with a discrete number of values for each.
    Can be used to iterate over parameter value combinations with the
    Python built-in function iter.

    Read more in the :ref:`User Guide <grid_search>`.
    Parameters
    ----------
    param_grid : dict of string to sequence, or sequence of such
        The parameter grid to explore, as a dictionary mapping estimator
        parameters to sequences of allowed values.
        An empty dict signifies default parameters.
        A sequence of dicts signifies a sequence of grids to search, and is
        useful to avoid exploring parameter combinations that make no sense
        or have no effect. See the examples below.
    Examples
    --------
    >>> from sklearn.grid_search import ParameterGrid
    >>> param_grid = {'a': [1, 2], 'b': [True, False]}
    >>> list(ParameterGrid(param_grid)) == (
    ...    [{'a': 1, 'b': True}, {'a': 1, 'b': False},
    ...     {'a': 2, 'b': True}, {'a': 2, 'b': False}])
    True
    >>> grid = [{'kernel': ['linear']}, {'kernel': ['rbf'], 'gamma': [1, 10]}]
    >>> list(ParameterGrid(grid)) == [{'kernel': 'linear'},
    ...                               {'kernel': 'rbf', 'gamma': 1},
    ...                               {'kernel': 'rbf', 'gamma': 10}]
    True
    >>> ParameterGrid(grid)[1] == {'kernel': 'rbf', 'gamma': 1}
    True
    """

    def __init__(self, param_grid):
        if isinstance(param_grid, Mapping):
            # wrap dictionary in a singleton list to support either dict
            # or list of dicts
            param_grid = [param_grid]
        self.param_grid = param_grid

    def __iter__(self):
        """Iterate over the points in the grid.
        Returns
        -------
        params : iterator over dict of string to any
            Yields dictionaries mapping each estimator parameter to one of its
            allowed values.
        """
        for p in self.param_grid:
            # Always sort the keys of a dictionary, for reproducibility
            items = sorted(p.items())
            if not items:
                yield {}
            else:
                keys, values = zip(*items)
                for v in itertools.product(*values):
                    params = OrderedDict(zip(keys, v))
                    yield params

    def __len__(self):
        """Number of points on the grid."""
        # Product function that can handle iterables (np.product can't).
        product = partial(reduce, operator.mul)
        return sum(product(len(v) for v in p.values()) if p else 1
                   for p in self.param_grid)

    def __getitem__(self, ind):
        """Get the parameters that would be ``ind``th in iteration
        Parameters
        ----------
        ind : int
            The iteration index
        Returns
        -------
        params : dict of string to any
            Equal to list(self)[ind]
        """
        # This is used to make discrete sampling without replacement memory
        # efficient.
        for sub_grid in self.param_grid:
            # XXX: could memoize information used here
            if not sub_grid:
                if ind == 0:
                    return {}
                else:
                    ind -= 1
                    continue

            # Reverse so most frequent cycling parameter comes first
            keys, values_lists = zip(*sorted(sub_grid.items())[::-1])
            sizes = [len(v_list) for v_list in values_lists]
            product = partial(reduce, operator.mul)
            total = product(sizes)

            if ind >= total:
                # Try the next grid
                ind -= total
            else:
                out = {}
                for key, v_list, n in zip(keys, values_lists, sizes):
                    ind, offset = divmod(ind, n)
                    out[key] = v_list[offset]
                return out

        raise IndexError('ParameterGrid index out of range')




from .crumb  import Crumb
from .mktree import mktree
from .utils  import (ParameterGrid,
                     intersection,
                     difference,
                     groupby_pattern,
                     crumb_copy,
                     crumb_link)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Crumb manipulation utilities
"""
import functools
import os
import os.path as op
import warnings
from string import Formatter

from   six import string_types

_txt_idx = 0
_fld_idx = 1
_rgx_idx = 2
_cnv_idx = 3


def _yield_items(crumb_path, index=None):
    """ An iterator over the items in `crumb_path` given by string.Formatter."""
    if index is None:
        return Formatter().parse(crumb_path)
    else:
        #for (literal_text, field_name, format_spec, conversion) in fmt.parse(crumb_path):
        # (txt, fld, fmt, conv)
        return (items[index] for items in Formatter().parse(crumb_path))


def _enum_items(crumb_path):
    """ An iterator over the enumerated items, i.e., (index, items) in
    `crumb_path` given by string.Formatter. """
    return ((idx, items) for idx, items in enumerate(Formatter().parse(crumb_path)))


def _depth_items(crumb_path, index=None):
    """ Return a generator with  (depth, items) in `crumb_path`. Being `depth`
     the place in the file path each argument is."""
    if index is None:
        index = slice(_txt_idx, _cnv_idx+1)

    depth = 0
    for idx, items in _enum_items(crumb_path):
        if items[_fld_idx]:
            depth += items[_txt_idx].count(op.sep)
            yield depth, items[index]


def _arg_names(crumb_path):
    """ Return an iterator over arg_name in crumb_path."""
    return _yield_items(crumb_path, _fld_idx)


def _depth_names(crumb_path):
    """ Return an iterator over (depth, arg_name)."""
    return _depth_items(crumb_path, _fld_idx)


def _depth_names_regexes(crumb_path):
    """ Return an iterator over (depth, (arg_name, arg_regex))."""
    return _depth_items(crumb_path, slice(_fld_idx, _cnv_idx))


def _build_path(crumb_path, arg_values, with_regex=True, regexes=None):
    """ Build the crumb_path with the values in arg_values.
    Parameters
    ----------
    crumb_path: str

    arg_values: dict[str]->str
        arg_name -> arg_value

    with_regex: bool

    regexes: dict[str] -> str
        dict[arg_name] -> regex
        The regexes contained here will replace or be added as a regex for
        the corresponding arg_name.

    Returns
    -------
    built_path: str
    """
    if regexes is None:
        regexes = {}

    path = ''
    for txt, fld, rgx, conv in _yield_items(crumb_path):
        path += txt
        if fld is None:
            continue

        if fld in arg_values:
            path += arg_values[fld]
        else:
            regex = regexes.get(fld, rgx) if with_regex else ''
            path += _format_arg(fld, regex=regex)

    return path


def is_valid(crumb_path):
    """ Return True if `crumb_path` is a valid Crumb value, False otherwise. """
    try:
        _ = list(_depth_names_regexes(crumb_path))
    except ValueError:
        return False
    else:
        return True


def _first_txt(crumb_path):
    """ Return the first text part without arguments in `crumb_path`. """
    for txt in _yield_items(crumb_path, index=_txt_idx):
        return txt


def _find_arg_depth(crumb_path, arg_name):
    """ Return the depth, name and regex of the argument with name `arg_name`.
    """
    for depth, (txt, fld, rgx, conv) in _depth_items(crumb_path):
        if fld == arg_name:
            return depth, fld, rgx


def _has_arg(crumb_path, arg_name):
    """ Return the True if the `arg_name` is found in `crumb_path`. """
    for txt, fld, rgx, conv in _yield_items(crumb_path):
        if fld == arg_name:
            return True
    return False


def _check(crumb_path):
    """ Raises some Errors if there is something wrong with `crumb_path`, if
    not the type needed or is not valid.
    Parameters
    ----------
    crumb_path: str

    Raises
    ------
     - ValueError if the path of the Crumb has errors using `self.is_valid`.
     - TypeError if the crumb_path is not a str or a Crumb.
    """
    if not isinstance(crumb_path, string_types):
        raise TypeError("Expected `crumb_path` to be a {}, "
                        "got {}.".format(string_types, type(crumb_path)))

    if not is_valid(crumb_path):
        raise ValueError("The current crumb path has errors, "
                         "got {}.".format(crumb_path))

    return crumb_path


def _get_path(crumb_path):
    """ Return the path string from `crumb_path`.
    Parameters
    ----------
    crumb_path: str or Crumb

    Returns
    -------
    path: str
    """
    if hasattr(crumb_path, '_path'):
        crumb_path = crumb_path._path

    if not isinstance(crumb_path, string_types):
        raise TypeError("Expected `crumb_path` to be a {}, "
                        "got {}.".format(string_types, type(crumb_path)))

    return crumb_path


def _is_crumb_arg(crumb_arg):
    """ Return True if `crumb_arg` is a well formed crumb argument, i.e.,
    is a string that starts with `start_sym` and ends with `end_sym`.
    False otherwise.
    """
    if not isinstance(crumb_arg, string_types):
        return False
    start_sym, end_sym = ('{', '}')
    return crumb_arg.startswith(start_sym) and crumb_arg.endswith(end_sym)


def _format_arg(arg_name, regex=''):
    """ Return the crumb argument for its string `format()` representation.
    Parameters
    ----------
    arg_name: str

    Returns
    -------
    arg_format: str
    """
    start_sym, end_sym = ('{', '}')
    reg_sym = ':'

    arg_fmt = start_sym + arg_name
    if regex:
        arg_fmt += reg_sym + regex
    arg_fmt += end_sym

    return arg_fmt


def has_crumbs(crumb_path):
    """ Return True if the `crumb_path.split(op.sep)` has item which is a
    crumb argument that starts with '{' and ends with '}'."""
    crumb_path = _get_path(crumb_path)

    splt = crumb_path.split(op.sep)
    for i in splt:
        if _is_crumb_arg(i):
            return True

    return False


def _split(crumb_path):
    """ Split `crumb_path` in two parts, the first is the base folder without
        any crumb argument and the second is the rest of `crumb_path` beginning
        with the first crumb argument.
        If `crumb_path` starts with an argument, will return ('', crumb_path).
    """
    crumb_path = _get_path(crumb_path)

    if not has_crumbs(crumb_path):
        return crumb_path, ''

    if not is_valid(crumb_path):
        raise ValueError('Crumb path {} is not valid.'.format(crumb_path))

    start_sym, end_sym = '{', '}'
    if crumb_path.startswith(start_sym):
        base = ''
        rest = crumb_path
    else:
        idx = crumb_path.find(start_sym)
        base = crumb_path[0:idx]
        if base.endswith(op.sep):
            base = base[:-1]

        rest = crumb_path[idx:]

    return base, rest


def _makedirs(dirpath, exist_ok=True):
    """ Is a replacement for os.makedirs for Python 2.7, with the `exist_ok`
    argument."""
    if op.exists(dirpath):
        if not exist_ok:
            raise IOError("Folder {} already exists.".format(dirpath))
        else:
            return dirpath

    return _make_new_dirs(dirpath)


def _make_new_dirs(dirpath):
    """ Call os.makedirs(dirpath), will return dirpath if no exception is
    raised."""
    try:
        os.makedirs(dirpath)
    except:
        raise
    else:
        return dirpath


def _touch(crumb_path, exist_ok=True):
    """ Create a leaf directory and all intermediate ones
    using the non crumbed part of `crumb_path`.
    If the target directory already exists, raise an IOError
    if exist_ok is False. Otherwise no exception is raised.
    Parameters
    ----------
    crumb_path: str

    exist_ok: bool
        Default = True

    Returns
    -------
    nupath: str
        The new path created.
    """
    if has_crumbs(crumb_path):
        nupath = _split(crumb_path)[0]
    else:
        nupath = crumb_path

    return _makedirs(nupath, exist_ok=exist_ok)


def _split_exists(crumb_path):
    """ Return True if the part without crumb arguments of `crumb_path`
    is an existing path or a symlink, False otherwise.
    Returns
    -------
    exists: bool
    """
    if has_crumbs(crumb_path):
        rpath = _split(crumb_path)[0]
    else:
        rpath = str(crumb_path)

    return op.exists(rpath) or op.islink(rpath)


def _check_is_subset(list1, list2):
    """ Raise an error if `list1` is not a subset of `list2`."""
    if not set(list1).issubset(set(list2)):
        raise KeyError('The `list1` argument should be a subset of `list2`, '
                       'got {} and {}.'.format(list1, list2))


def deprecated(replacement=None):
    """A decorator which can be used to mark functions as deprecated.
    replacement is a callable that will be called with the same args
    as the decorated function.

    >>> @deprecated()
    ... def foo(x):
    ...     return x
    ...
    >>> ret = foo(1)
    DeprecationWarning: foo is deprecated
    >>> ret
    1
    >>>
    >>>
    >>> def newfun(x):
    ...     return 0
    ...
    >>> @deprecated(newfun)
    ... def foo(x):
    ...     return x
    ...
    >>> ret = foo(1)
    DeprecationWarning: foo is deprecated; use newfun instead
    >>> ret
    0
    >>>
    """
    def outer(fun):
        msg = "psutil.%s is deprecated" % fun.__name__
        if replacement is not None:
            msg += "; use %s instead" % replacement
        if fun.__doc__ is None:
            fun.__doc__ = msg

        @functools.wraps(fun)
        def inner(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
            return fun(*args, **kwargs)

        return inner
    return outer



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Crumb class: the smart path model class.
"""

from collections import Mapping


def mktree(crumb, values_map):
    """ Create the tree of folders given the values for the crumb arguments
    of the current crumb path.
    Parameters
    ----------
    crumb: Crumb

    values_map: Sequence[Sequence[2-Tuple[str, str]]] or Sequence[Dict[str, str]]
        The lists of values to substitute each crumb argument that you want.
        Do not leave dependent arguments alone.
        Example: [[('subject_id', 'pat_0001'), ('session_id', 'session_1')],
                  [('subject_id', 'pat_0002'), ('session_id', 'session_1')],
                  ....
                 ]

        Example: [{'subject_id': 'pat_0001', 'session_id': 'session_1'},
                  {'subject_id': 'pat_0002', 'session_id': 'session_1'},
                  ....
                 ]

    Returns
    -------
    paths: list of Paths
        The paths that have been created.
    """
    if values_map is None:
        return [crumb.touch()]

    if not isinstance(values_map, (list, dict)):
        raise TypeError("Expected keys in `values_map` to be a Sequence, "
                         "got {}.".format(type(values_map)))

    paths = []
    for idx, aval in enumerate(values_map):
        if not isinstance(aval, Mapping):
            aval = dict(aval)

        if not set(aval.keys()).issubset(set(crumb.all_args())):
            raise ValueError("Expected keys in `values_map` item to be a subset of {}, "
                             "got {}.".format(crumb.all_args(), aval.keys()))

        rem_deps = crumb._args_open_parents(list(aval.keys()))
        if rem_deps:
            raise KeyError("Expected `values_map` item to not leave crumbs alone,"
                           " you forgot to add: {} in item {}".format(rem_deps, idx))

        paths.append(crumb.replace(**aval))

    _ = [path.touch() for path in paths]

    return paths



#!python
import click

from .utils import (CONTEXT_SETTINGS,
                    CrumbPath,
                    echo_list,
                    _print_values_map_as_csv,
                    check_not_none,)


# declare the CLI group
@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('crumb', type=CrumbPath(), callback=check_not_none)
@click.option('-i', '--ignore', type=str, multiple=True,
              help='A global ignore fnmatch expression for the listing. '
                   'You can add as many of this argument as you want. '
                   'Example: ".*" or "*~"')
@click.option('-a', '--arg', type=str,
              help='Name of the argument in `crumb` to print the values from.'
                   'Will not print full paths, but only values for this '
                   'crumb argument.')
def ls(crumb, ignore, arg):
    """List all the possible values that match the given crumb path.

    Examples: \n
    crumb ls "/data/hansel/cobre/{sid:4*100}/{session}/{img}"\n
    crumb ls -i ".*" "/data/hansel/cobre/{sid}/{session}/{img:anat*}"\n
    crumb ls -a "sid" "/data/hansel/cobre/{sid}/{session}/{img:anat*}"\n
    """
    if not crumb.isabs():
        crumb = crumb.abspath()

    crumb._ignore = ignore
    if arg:
        lst = crumb[arg]
    else:
        lst = crumb.ls()

    if not lst:
        exit(-1)

    echo_list(lst)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('src_crumb', type=CrumbPath(), callback=check_not_none)
@click.argument('dst_crumb', type=CrumbPath(), callback=check_not_none)
@click.option('-q', '--quiet', is_flag=True, flag_value=True,
              help='Flag to remove verbose.')
@click.option('-e', '--exist_ok', is_flag=True, flag_value=True,
              help='Flag to allow overwriting destination path.')
@click.option('-i', '--ignore', type=str, multiple=True,
              help='A global ignore fnmatch expression for the listing. '
                   'You can add as many of this argument as you want. '
                   'Example: ".*" or "*~"')
def copy(src_crumb, dst_crumb, quiet, ignore, exist_ok):
    """Copy one file tree to another file tree. The
    structure of the destination tree can be modified.

    Examples: \n
    crumb copy "/data/hansel/cobre/{sid}/{session}/{img}" "/data/hansel/cobre2/{sid}/{img}" \n
    crumb copy "cobre/{sid}/{session}/{img:anat*}" "cobre_anat/{sid}/{img}" \n
    """
    from .. import crumb_copy

    if ignore:
        src_crumb._ignore = ignore
        dst_crumb._ignore = ignore

    if not src_crumb.ls():
        click.echo('Could not find any file that matched {}.'.format(src_crumb))
        exit(-1)

    crumb_copy(src_crumb, dst_crumb,
               exist_ok=exist_ok,
               verbose=(not quiet))



@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('src_crumb', type=CrumbPath(), callback=check_not_none)
@click.argument('dst_crumb', type=CrumbPath(), callback=check_not_none)
@click.option('-q', '--quiet', is_flag=True, flag_value=True,
              help='Flag to remove verbose.')
@click.option('-e', '--exist_ok', is_flag=True, flag_value=True,
              help='Flag to allow overwriting destination path.')
@click.option('-i', '--ignore', type=str, multiple=True,
              help='A global ignore fnmatch expression for the listing. '
                   'You can add as many of this argument as you want. '
                   'Example: ".*" or "*~"')
def link(src_crumb, dst_crumb, quiet, ignore, exist_ok):
    """Link one file tree to another file tree. The
    structure of the destination tree can be modified.
    Only the leaf nodes will be linked, the folder structure above will be
    created.

    Examples: \n
    crumb link "/data/hansel/cobre/{sid}/{session}/{img}" "/data/hansel/cobre2/{sid}/{img}" \n
    crumb link "cobre/{sid}/{session}/{img:anat*}" "cobre_anat/{sid}/{img}" \n
    """
    from .. import crumb_link

    if ignore:
        src_crumb._ignore = ignore
        dst_crumb._ignore = ignore

    if not src_crumb.ls():
        click.echo('Could not find any file that matched {}.'.format(src_crumb))
        exit(-1)

    crumb_link(src_crumb, dst_crumb,
               exist_ok=exist_ok,
               verbose=(not quiet))


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('crumb1', type=CrumbPath(), callback=check_not_none)
@click.argument('crumb2', type=CrumbPath(), callback=check_not_none)
@click.option('-o', '--on', type=str, multiple=True,
              help='Argument name to check for intersection. You can use this '
                   'argument more than once.')
@click.option('-b', '--base', type=click.Choice(['1', '2']), default='0',
              help='1 or 2, to indicate `crumb1` or `crumb2` as a base crumb2 '
                   'to print the results')
def intersect(crumb1, crumb2, on, base):
    """Return the intersection between crumb1 and crumb2 on a given argument.

    Will not print full paths, but only values for the crumb arguments in `on`.
    Unless you specify the crumb you want to use as a base with the `base`
    argument.

    Examples: \n
    crumb intersect --on "sid" "/data/hansel/cobre/{sid}/{session}/{img}" "/data/hansel/cobre2/{sid}/{img}"\n
    """
    from .. import intersection

    values = intersection(crumb1, crumb2, on=on)
    if base == '0':
        _print_values_map_as_csv(values)
    else:
        if base == '1':
            base_crumb = crumb1
        else:
            base_crumb = crumb2

        echo_list(base_crumb.build_paths(values))


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('crumb1', type=CrumbPath(), callback=check_not_none)
@click.argument('crumb2', type=CrumbPath(), callback=check_not_none)
@click.option('-o', '--on', type=str, multiple=True,
              help='Argument name to check for intersection. You can use this '
                   'argument more than once.')
@click.option('-b', '--base', type=click.Choice(['1', '2']), default='0',
              help='1 or 2, to indicate `crumb1` or `crumb2` as a base crumb2 '
                   'to print the results')
def diff(crumb1, crumb2, on, base):
    """Return the difference crumb1 - crumb2 on a given argument.

    Will not print full paths, but only values for the crumb arguments in `on`.
    Unless you specify the crumb you want to use as a base with the `base`
    argument.

    Examples: \n
    crumb diff --on "sid" "/data/hansel/cobre/{sid}/{session}/{img}" "/data/hansel/cobre2/{sid}/{img}"\n
    """
    from .. import difference

    values = difference(crumb1, crumb2, on=on)
    if base == '0':
        _print_values_map_as_csv(values)
    else:
        if base == '1':
            base_crumb = crumb1
        else:
            base_crumb = crumb2

        echo_list(base_crumb.build_paths(values))



#!/usr/bin/env python

"""
hansel
------

Flexible parametric file paths to make queries, build folder trees and
smart folder structure access.
"""

from __future__ import print_function

import os.path as op
import io
import sys

from   setuptools              import setup, find_packages
from   setuptools.command.test import test as TestCommand


# long description
def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


# Get version without importing, which avoids dependency issues
MODULE_NAME = find_packages(exclude=['tests'])[0]
VERSION_PYFILE = op.join(MODULE_NAME, 'version.py')
# set __version__ variable
exec(compile(read(VERSION_PYFILE), VERSION_PYFILE, 'exec'))


setup_dict = dict(
    name=MODULE_NAME,
    version=__version__,
    description='Easily traverse your structured folder tree.',

    url='https://pypi.python.org/pypi/hansel',
    license='Apache 2.0',
    author='alexsavio',
    author_email='alexsavio@gmail.com',
    maintainer='alexsavio',
    maintainer_email='alexsavio@gmail.com',

    packages=find_packages(),

    install_requires=['six'],

    entry_points='''
      [console_scripts]
      crumb=hansel.cli:cli
      ''',

    long_description=read('README.rst', 'CHANGES.rst'),

    platforms='Linux/MacOSX',

    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 2.7',
    ],

    extras_require={
        'testing': ['pytest', 'pytest-cov', 'pandas'],
    }
)


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup_dict.update(dict(tests_require=['pytest'],
                       cmdclass={'test': PyTest}))


if __name__ == '__main__':
    setup(**setup_dict)
